/* SPDX-License-Identifier: GPL-2.0 */

#ifndef _ASM_X86_CPU_ENTRY_AREA_H
#define _ASM_X86_CPU_ENTRY_AREA_H

#include <linux/percpu-defs.h>
#include <asm/processor.h>
#include <asm/intel_ds.h>

#ifdef CONFIG_X86_64

/* Macro to enforce the same ordering and stack sizes */
#define ESTACKS_MEMBERS(guardsize, db2_holesize)\
	char	DF_stack_guard[guardsize];	\
	char	DF_stack[EXCEPTION_STKSZ];	\
	char	NMI_stack_guard[guardsize];	\
	char	NMI_stack[EXCEPTION_STKSZ];	\
	char	DB2_stack_guard[guardsize];	\
	char	DB2_stack[db2_holesize];	\
	char	DB1_stack_guard[guardsize];	\
	char	DB1_stack[EXCEPTION_STKSZ];	\
	char	DB_stack_guard[guardsize];	\
	char	DB_stack[EXCEPTION_STKSZ];	\
	char	MCE_stack_guard[guardsize];	\
	char	MCE_stack[EXCEPTION_STKSZ];	\
	char	IST_top_guard[guardsize];	\

/* The exception stacks' physical storage. No guard pages required */
struct exception_stacks {
	ESTACKS_MEMBERS(0, 0)
};

/* The effective cpu entry area mapping with guard pages. */
struct cea_exception_stacks {
	ESTACKS_MEMBERS(PAGE_SIZE, EXCEPTION_STKSZ)
};

/*
 * The exception stack ordering in [cea_]exception_stacks
 */
enum exception_stack_ordering {
	ESTACK_DF,
	ESTACK_NMI,
	ESTACK_DB2,
	ESTACK_DB1,
	ESTACK_DB,
	ESTACK_MCE,
	N_EXCEPTION_STACKS
};

#define CEA_ESTACK_SIZE(st)					\
	sizeof(((struct cea_exception_stacks *)0)->st## _stack)

#define CEA_ESTACK_BOT(ceastp, st)				\
	((unsigned long)&(ceastp)->st## _stack)

#define CEA_ESTACK_TOP(ceastp, st)				\
	(CEA_ESTACK_BOT(ceastp, st) + CEA_ESTACK_SIZE(st))

#define CEA_ESTACK_OFFS(st)					\
	offsetof(struct cea_exception_stacks, st## _stack)

#define CEA_ESTACK_PAGES					\
	(sizeof(struct cea_exception_stacks) / PAGE_SIZE)

#endif

/*
 * cpu_entry_area is a percpu region that contains things needed by the CPU
 * and early entry/exit code.  Real types aren't used for all fields here
 * to avoid circular header dependencies.
 *
 * Every field is a virtual alias of some other allocated backing store.
 * There is no direct allocation of a struct cpu_entry_area.
 */
struct cpu_entry_area {
	char gdt[PAGE_SIZE];

	/*
	 * The GDT is just below entry_stack and thus serves (on x86_64) as
	 * a read-only guard page. On 32-bit the GDT must be writeable, so
	 * it needs an extra guard page.
	 */
#ifdef CONFIG_X86_32
	char guard_entry_stack[PAGE_SIZE];
#endif
	struct entry_stack_page entry_stack_page;

	/*
	 * On x86_64, the TSS is mapped RO.  On x86_32, it's mapped RW because
	 * we need task switches to work, and task switches write to the TSS.
	 */
	struct tss_struct tss;

#ifdef CONFIG_X86_64
	/*
	 * Exception stacks used for IST entries with guard pages.
	 */
	struct cea_exception_stacks estacks;
#endif
	/*
	 * Per CPU debug store for Intel performance monitoring. Wastes a
	 * full page at the moment.
	 */
	struct debug_store cpu_debug_store;
	/*
	 * The actual PEBS/BTS buffers must be mapped to user space
	 * Reserve enough fixmap PTEs.
	 */
	struct debug_store_buffers cpu_debug_buffers;
};

#define CPU_ENTRY_AREA_SIZE		(sizeof(struct cpu_entry_area))

DECLARE_PER_CPU(struct cpu_entry_area *, cpu_entry_area);
DECLARE_PER_CPU(struct cea_exception_stacks *, cea_exception_stacks);

extern void setup_cpu_entry_areas(void);
extern void cea_set_pte(void *cea_vaddr, phys_addr_t pa, pgprot_t flags);

/* Single page reserved for the readonly IDT mapping: */
#define	CPU_ENTRY_AREA_RO_IDT		CPU_ENTRY_AREA_BASE
#define CPU_ENTRY_AREA_PER_CPU		(CPU_ENTRY_AREA_RO_IDT + PAGE_SIZE)

#define CPU_ENTRY_AREA_RO_IDT_VADDR	((void *)CPU_ENTRY_AREA_RO_IDT)

#ifdef CONFIG_X86_32
#define CPU_ENTRY_AREA_MAP_SIZE		(CPU_ENTRY_AREA_PER_CPU +		\
					 (CPU_ENTRY_AREA_SIZE * NR_CPUS) -	\
					 CPU_ENTRY_AREA_BASE)
#else
#define CPU_ENTRY_AREA_MAP_SIZE		P4D_SIZE
#endif

extern struct cpu_entry_area *get_cpu_entry_area(int cpu);

static inline struct entry_stack *cpu_entry_stack(int cpu)
{
	return &get_cpu_entry_area(cpu)->entry_stack_page.stack;
}

#define __this_cpu_ist_top_va(name)					\
	CEA_ESTACK_TOP(__this_cpu_read(cea_exception_stacks), name)

#define __this_cpu_ist_bottom_va(name)					\
	CEA_ESTACK_BOT(__this_cpu_read(cea_exception_stacks), name)


/* HYGON added to avoid kabi breakage of struct cpu_entry_area (start) */

#ifdef CONFIG_X86_64

#define ESTACKS_MEMBERS_HYGON(guardsize, optional_stack_size)		\
	char	VC_stack_guard[guardsize];				\
	char	VC_stack[optional_stack_size];				\
	char	VC2_stack_guard[guardsize];				\
	char	VC2_stack[optional_stack_size];				\
	char	IST_top_guard[guardsize];				\

/* The effective VC cpu entry area mapping with guard pages. */
struct cea_exception_stacks_hygon {
	ESTACKS_MEMBERS_HYGON(PAGE_SIZE, EXCEPTION_STKSZ)
};

enum exception_stack_ordering_hygon {
	ESTACK_VC = N_EXCEPTION_STACKS,
	ESTACK_VC2,
	N_EXCEPTION_STACKS_HYGON
};

#define CEA_ESTACK_HYGON_SIZE(st)				\
	sizeof(((struct cea_exception_stacks_hygon *)0)->st## _stack)

#define CEA_ESTACK_HYGON_BOT(ceahygonstp, st)			\
	((unsigned long)&(ceahygonstp)->st## _stack)

#define CEA_ESTACK_HYGON_TOP(ceahygonstp, st)			\
	(CEA_ESTACK_HYGON_BOT(ceahygonstp, st) + CEA_ESTACK_HYGON_SIZE(st))

#define CEA_ESTACK_HYGON_OFFS(st)				\
	offsetof(struct cea_exception_stacks_hygon, st## _stack)

#define CEA_ESTACK_HYGON_PAGES					\
	(sizeof(struct cea_exception_stacks) / PAGE_SIZE)

struct cpu_entry_area_hygon {
	/*
	 * Exception stacks used for HYGON CSV2 IST entries with guard pages.
	 */
	struct cea_exception_stacks_hygon estacks;
};

#define CPU_ENTRY_AREA_HYGON_SIZE	(sizeof(struct cpu_entry_area_hygon))

DECLARE_PER_CPU(struct cpu_entry_area_hygon *, cpu_entry_area_hygon);
DECLARE_PER_CPU(struct cea_exception_stacks_hygon *, cea_exception_stacks_hygon);

#define cea_set_pte_hygon(vaddr, pa, flags)	cea_set_pte(vaddr, pa, flags)

extern struct cpu_entry_area_hygon *get_cpu_entry_area_hygon(int cpu);

#define __this_cpu_ist_top_va_hygon(name)					\
	CEA_ESTACK_HYGON_TOP(__this_cpu_read(cea_exception_stacks_hygon), name)

#define __this_cpu_ist_bottom_va_hygon(name)					\
	CEA_ESTACK_HYGON_BOT(__this_cpu_read(cea_exception_stacks_hygon), name)

#else /* !CONFIG_X86_64 */

#define CPU_ENTRY_AREA_HYGON_SIZE	0

#endif	/* CONFIG_X86_64 */

/* HYGON added to avoid kabi breakage of struct cpu_entry_area (end) */

#endif
