#!/bin/bash

set -e

LD=ld
DKMS_BUILD_DIR="$( cd -- "$(dirname "$0")" > /dev/null 2>&1 ; pwd -P)"
PROJECT_DIR=$DKMS_BUILD_DIR/../..
TARGET=$PROJECT_DIR/dkms_target
DKMS_VERSION=${1:-'1.0.0'}

FIND_FILTER=("-not" "-path" "*rogue/*" \
	"-not" "-path" "*innosrvkm/bridge/*" \
	"-not" "-path" "*rgx_nohw/*" \
	"-not" "-path" "*rgx_linux_plato/*" \
	"-not" "-path" "*rgx_linux_tc/*")

OPEN_SRCS_SUFFIX=_open_srcs.txt
BIN_EXCLUDE_OBJ_SUFFIX=_bin_exclude_objs.tmp

function prepare_env_and_make() {
	cd $PROJECT_DIR
	make -j16
	make export_config
	sed -i 's/^/export /' $DKMS_BUILD_DIR/modules_config.sh
	source $DKMS_BUILD_DIR/modules_config.sh

	if [[ -n $CROSS_COMPILE ]]; then
		LD=${CROSS_COMPILE}ld
	fi
}

function generate_kbuild() {

	if [[ $CONFIG_DRM_INNO_VPU != "y" ]]; then
		sed -i '/innovpu/Id' $TARGET/Kbuild
	fi

	if [[ $CONFIG_DRM_INNO_DPU != "y" ]]; then
		sed -i '/innodpu/Id' $TARGET/Kbuild
	fi

	if [[ $CONFIG_INNO_ION != "y" ]]; then
		sed -i '/innoion/Id' $TARGET/Kbuild
	fi

	if [[ $CONFIG_DRM_INNO_PMBUS != "y" ]]; then
		sed -i '/innopmbus/Id' $TARGET/Kbuild
	fi

	if [[ $CONFIG_DRM_INNO_AUDIO != "y" ]]; then
		sed -i '/innoaudio/Id' $TARGET/Kbuild
	fi

	if [[ $CONFIG_DRM_INNO_POWER != "y" ]]; then
		sed -i '/innopower/Id' $TARGET/Kbuild
	fi

	if [[ $CONFIG_DRM_INNO_SR != "y" ]]; then
		sed -i '/innosr[^v]/Id' $TARGET/Kbuild
	fi

	if [[ $CONFIG_DRM_INNO_DEBUG != "y" ]]; then
		sed -i '/tools/Id' $TARGET/Kbuild
		sed -i '/innogpu_info/Id' $TARGET/Kbuild
	fi

	sed -i 's/$(CONFIG_DRM_INNO_GPU)/m/' $TARGET/Kbuild
}

# 模块是否有闭源的二进制
function module_has_private_objs() {
	module=$1
	objs_dir=$PROJECT_DIR/${module}/
	bin_exclude_objs=$DKMS_BUILD_DIR/${module}${BIN_EXCLUDE_OBJ_SUFFIX}

	find $objs_dir -name "*.o" | grep -vFf $bin_exclude_objs | wc -l
}

# 生成链接二进制需要排除的.o文件
# 该函数是通用实现,不要轻易修改,如果模块有特殊需求自行实现
function prepare_bin_exclude_objs() {
	module=$1
	open_srcs=$DKMS_BUILD_DIR/${module}${OPEN_SRCS_SUFFIX}
	bin_exclude_objs=$DKMS_BUILD_DIR/${module}${BIN_EXCLUDE_OBJ_SUFFIX}

	cat $open_srcs | grep "\.c" > $bin_exclude_objs
	sed -i 's/\.c/.o/' $bin_exclude_objs
}

# 生成编译开源模块的makefile文件
# 该函数是通用实现,不要轻易修改,如果模块有特殊需求自行实现
function generate_makefile() {
	module=$1
	makefile=$TARGET/${module}/Makefile
	cnt=$(module_has_private_objs $module)
	cp $DKMS_BUILD_DIR/${module}${BIN_EXCLUDE_OBJ_SUFFIX} $makefile

	sed -i 's/^/'"${module}"'\//' $makefile
	test -s $makefile && sed -i '1i\'"${module^^}"'_OBJS =' $makefile || \
		echo "${module^^}_OBJS =" >> $makefile
	sed -i 's/$/ \\/' $makefile
	if [[ $cnt != 0 ]]; then
		sed -i '$a\'"${module}"'/'"${module}"'.o' $makefile
	fi
	sed -i '$a\\nifeq ($(SEPARATE_BUILD),y)' $makefile
	sed -i '$a\  '"${module}"'-y := $('"${module^^}"'_OBJS)' $makefile
	sed -i '$a\  obj-m += '"${module}"'.o' $makefile
	sed -i '$a\else' $makefile
	sed -i '$a\  '"${module^^}"' := $('"${module^^}"'_OBJS)' $makefile
	sed -i '$a\endif' $makefile
	sed -i '1i\'"${module^^}"'_INCLUDE :=-I$(INNOGPU_PRO_PATH)/'"${module}"'' $makefile
}

# 生成二进制
# 该函数是通用实现,如果模块有特殊需求自行实现,参照innosrvkm
function generate_bin() {
	module=$1
	objs_dir=$PROJECT_DIR/${module}/
	bin_name=$TARGET/${module}/${module}.o_shipped
	bin_exclude_objs=$DKMS_BUILD_DIR/${module}${BIN_EXCLUDE_OBJ_SUFFIX}
	cnt=$(module_has_private_objs $module)

	if [[ $cnt != 0 ]]; then
		$LD -r -o $bin_name `find $objs_dir -name "*.o" | \
			grep -vFf $bin_exclude_objs`
	fi
}

# 拷贝开源的代码
# 该函数是通用实现,如果模块有特殊需求自行实现,参照innosrvkm
function copy_open_srcs() {
	module=$1
	src_dir=$PROJECT_DIR/${module}
	dst_dir=$TARGET/${module}
	open_srcs=$DKMS_BUILD_DIR/${module}${OPEN_SRCS_SUFFIX}

	mkdir -p $dst_dir

	for i in $(cat $open_srcs); do
		find $src_dir -name $i -exec cp -v {} $dst_dir \;
	done
}

function prepare_innogpu_exclude_objs() {
	open_srcs=$DKMS_BUILD_DIR/innogpu${OPEN_SRCS_SUFFIX}
	bin_exclude_objs=$DKMS_BUILD_DIR/innogpu${BIN_EXCLUDE_OBJ_SUFFIX}

	if [[ $CONFIG_INNO_ION != "y" ]]; then
		sed '/ionn*/d' $open_srcs | grep "\.c" > $bin_exclude_objs
	else
		cat $open_srcs | grep "\.c" > $bin_exclude_objs
	fi

	sed -i 's/\.c/.o/' $bin_exclude_objs
}

function prepare_innoaudio_exclude_objs() {
	prepare_bin_exclude_objs innoaudio
}

function prepare_innodma_exclude_objs() {
	prepare_bin_exclude_objs innodma
}

function prepare_innodpu_exclude_objs() {
	prepare_bin_exclude_objs innodpu
}

function prepare_innoion_exclude_objs() {
	prepare_bin_exclude_objs innoion
}

function prepare_innopmbus_exclude_objs() {
	prepare_bin_exclude_objs innopmbus
}

function prepare_innovpu_exclude_objs() {
	prepare_bin_exclude_objs innovpu
}

function prepare_innosr_exclude_objs() {
	prepare_bin_exclude_objs innosr
}

function prepare_innopower_exclude_objs() {
	prepare_bin_exclude_objs innopower
}

function prepare_innosrvkm_exclude_objs() {
	open_srcs=$DKMS_BUILD_DIR/innosrvkm_open_srcs.txt
	bin_objs=$DKMS_BUILD_DIR/innosrvkm${BIN_EXCLUDE_OBJ_SUFFIX}
	_FILE=${open_srcs}.tmp
	cp $open_srcs $_FILE

	if [[ $PVR_BUILD_DIR = android_container ]]; then
		sed -i '/pvr_sync_ioctl_drm.c/d' $_FILE
	else
		sed -i '/pvr_sync_ioctl_dev.*/d' $_FILE
		sed -i '/ion_sys.*/d' $_FILE
	fi

	if [[ $CONFIG_DRM_INNO_POWER != "y" ]]; then
		sed -i '/pvr_dvfs_device.*/d' $_FILE
	fi

	cat $_FILE | grep "\.c" > $bin_objs
	sed -i 's/\.c/.o/' $bin_objs
}

function generate_makefile_innogpu() {
	generate_makefile innogpu

	# innogpu不需要独立编译的控制
	sed -i '/SEPARATE/,+3d' $TARGET/innogpu/Makefile
	sed -i '/endif/d' $TARGET/innogpu/Makefile
}

function generate_makefile_innoaudio() {
	generate_makefile innoaudio
}

function generate_makefile_innodma() {
	generate_makefile innodma
}

function generate_makefile_innodpu() {
	generate_makefile innodpu
}

function generate_makefile_innoion() {
	generate_makefile innoion
}

function generate_makefile_innopmbus() {
	generate_makefile innopmbus
}

function generate_makefile_innosrvkm() {
	generate_makefile innosrvkm
	sed -i '1d' $TARGET/innosrvkm/Makefile
	sed -i '1i\'"${module^^}"'_INCLUDE :=-I$(INNOGPU_PRO_PATH)/'"${module}"'/include' \
		$TARGET/innosrvkm/Makefile
}

function generate_makefile_innopower() {
	generate_makefile innopower
}

function generate_makefile_innovpu() {
	generate_makefile innovpu
}

function generate_makefile_innosr() {
	generate_makefile innosr
}

function copy_innogpu_open_srcs() {
	copy_open_srcs innogpu
}

function copy_innoaudio_open_srcs() {
	copy_open_srcs innoaudio
}

function copy_innodma_open_srcs() {
	copy_open_srcs innodma
}

function copy_innodpu_open_srcs() {
	copy_open_srcs innodpu
}

function copy_innoion_open_srcs() {
	copy_open_srcs innoion
	cp -r $PROJECT_DIR/innoion/include $TARGET/innoion/include
}

function copy_innopmbus_open_srcs() {
	copy_open_srcs innopmbus
}

function copy_innosrvkm_open_srcs() {
	mkdir -p $TARGET/innosrvkm && mkdir -p $TARGET/innosrvkm/include
	cp $PROJECT_DIR/innosrvkm/include/config_kernel.h $TARGET/innosrvkm/include

	for i in $(cat $DKMS_BUILD_DIR/innosrvkm${OPEN_SRCS_SUFFIX}.tmp)
	do
		if [[ $i =~ "/" ]]; then
			DIR=${i%/*}
			FILE=${i#*/}
			mkdir -p $TARGET/innosrvkm/include/$DIR
			find $PROJECT_DIR/innosrvkm -name $FILE "${FIND_FILTER[@]}"  -exec cp -v {} $TARGET/innosrvkm/include/$DIR \;
		elif [[ $i =~ ".c" ]]; then
			find $PROJECT_DIR/innosrvkm -name $i "${FIND_FILTER[@]}" -exec cp -v {} $TARGET/innosrvkm \;
		else
			find $PROJECT_DIR/innosrvkm -name $i "${FIND_FILTER[@]}" -exec cp -v {} $TARGET/innosrvkm/include \;
		fi
	done
}

function copy_innovpu_open_srcs() {
	copy_open_srcs innovpu
}

function copy_innosr_open_srcs() {
	copy_open_srcs innosr
}

function copy_innopower_open_srcs() {
	copy_open_srcs innopower
}

function generate_bin_innogpu() {
	generate_bin innogpu
}

function generate_bin_innoaudio() {
	generate_bin innoaudio
}

function generate_bin_innodma() {
	generate_bin innodma
}

function generate_bin_innodpu() {
	generate_bin innodpu
}

function generate_bin_innoion() {
	generate_bin innoion
}

function generate_bin_innopmbus() {
	generate_bin innopmbus
}

function generate_bin_innosrvkm() {
	generate_bin innosrvkm
}

function generate_bin_innovpu() {
	generate_bin innovpu
}

function generate_bin_innosr() {
	generate_bin innosr
}

function generate_bin_innopower() {
	generate_bin innopower
}

function make_module() {
	module=$1
	prepare_${module}_exclude_objs
	copy_${module}_open_srcs
	generate_makefile_${module}
	generate_bin_${module}
}

function make_all_modules() {
	mkdir -p $TARGET

	make_module innogpu
	make_module innosrvkm
	make_module innodma

	if [[ $CONFIG_DRM_INNO_VPU = "y" ]]; then
		make_module innovpu
	fi

	if [[ $CONFIG_DRM_INNO_SR = "y" ]]; then
		make_module innosr
	fi

	if [[ $CONFIG_DRM_INNO_DPU = "y" ]]; then
		make_module innodpu
	fi

	if [[ $CONFIG_INNO_ION = "y" ]]; then
		make_module innoion
	fi

	if [[ $CONFIG_DRM_INNO_PMBUS = "y" ]]; then
		make_module innopmbus
	fi

	if [[ $CONFIG_DRM_INNO_AUDIO = "y" ]]; then
		make_module innoaudio
	fi

	if [[ $CONFIG_DRM_INNO_POWER = "y" ]]; then
		make_module innopower
	fi

	if [[ $CONFIG_DRM_INNO_DEBUG = "y" ]]; then
		mkdir -p $TARGET/tools
		cp -r $PROJECT_DIR/tools/gpu_info $TARGET/tools
		rm -rf $TARGET/tools/gpu_info/*.o
		cp $PROJECT_DIR/tools/Makefile $TARGET/tools
	fi

	cp $DKMS_BUILD_DIR/Makefile $TARGET
	cp $PROJECT_DIR/Kbuild $TARGET
	cp $DKMS_BUILD_DIR/dkms.conf $TARGET
	mv $DKMS_BUILD_DIR/modules_config.sh $TARGET/
	cp $DKMS_BUILD_DIR/cfg_detect.sh $TARGET
	cp $DKMS_BUILD_DIR/test_item.sh $TARGET

	cp -r $DKMS_BUILD_DIR/kernel_api_support_test $TARGET/tools
	sed -i '/^LINK_TEST_DIR/c LINK_TEST_DIR="./tools/kernel_api_support_test"' $TARGET/cfg_detect.sh
}

function set_dkms_version() {
	version=$1
	sed -i "s/PACKAGE_VERSION=\"2.1\"/PACKAGE_VERSION=\"${DKMS_VERSION}\"/" $TARGET/dkms.conf
}

function add_dkms_module() {
	name=$1
	cnt=$2
	sed -i "/AUTOINSTALL/i\BUILT_MODULE_NAME[${cnt}]=\"${name}\"" $TARGET/dkms.conf
	sed -i "/AUTOINSTALL/i\DEST_MODULE_LOCATION[${cnt}]=\"/kernel/drivers/gpu/drm/innogpu/\"" $TARGET/dkms.conf
}

function generate_dkms_conf() {
	set_dkms_version $DKMS_VERSION
	if [[ $SEPARATE_BUILD != "y" ]]; then
		return 0
	fi

	cnt=1
	add_dkms_module "innodma" $cnt
	let cnt=$cnt+1

	add_dkms_module "innosrvkm" $cnt
	let cnt=$cnt+1

	if [[ $CONFIG_DRM_INNO_VPU = "y" ]]; then
		add_dkms_module "innovpu" $cnt
		let cnt=$cnt+1
	fi

	if [[ $CONFIG_DRM_INNO_SR = "y" ]]; then
		add_dkms_module "innosr" $cnt
		let cnt=$cnt+1
	fi

	if [[ $CONFIG_DRM_INNO_DPU = "y" ]]; then
		add_dkms_module "innodpu" $cnt
		let cnt=$cnt+1
	fi

	if [[ $CONFIG_INNO_ION = "y" ]]; then
		add_dkms_module "innoion" $cnt
		let cnt=$cnt+1
	fi

	if [[ $CONFIG_DRM_INNO_PMBUS = "y" ]]; then
		add_dkms_module "innopmbus" $cnt
		let cnt=$cnt+1
	fi

	if [[ $CONFIG_DRM_INNO_AUDIO = "y" ]]; then
		add_dkms_module "innoaudio" $cnt
		let cnt=$cnt+1
	fi

	if [[ $CONFIG_DRM_INNO_POWER = "y" ]]; then
		add_dkms_module "innopower" $cnt
		let cnt=$cnt+1
	fi

	if [[ $CONFIG_DRM_INNO_DEBUG = "y" ]]; then
		add_dkms_module "innogpu_info" $cnt
		let cnt=$cnt+1
	fi
}

function make_dkms_deb() {
	cd $TARGET
	dkms mkdeb --source-only
}

# 清理
function clean() {
	rm -rf $DKMS_BUILD_DIR/*.tmp
	cd $PROJECT_DIR
	make clean
}

prepare_env_and_make
make_all_modules
generate_kbuild
generate_dkms_conf
make_dkms_deb
clean
